#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

import logging
from keysql_client.key_object import KeyObject, KEYOBJECT_TYPE
from keysql_client.key_object import KeyObjectInteger, KeyObjectNumber, KeyObjectChar, KeyObjectDate
from keysql_client.key_object import KeyObjectMultiple, KeyObjectComposition
from keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage
import json
from collections import OrderedDict

'''
class CreateKeyObjectStatement(KeySqlStatement):
    def __init__(self, keyobject, catalog_name):
        self._keyobject = keyobject
        self.catalog_name = catalog_name

    def create_request(self):
        return 'create keyobject {} in catalog {}'.format(self._keyobject.serialize(), self.catalog_name)
        sg = 'create keyobject ' + kn.serialize() + ' in catalog ' + catalog_name_new
'''

class CreateKeyObjectStatement(KeySqlStatement):
    """Base class for all Statement classes
    """
    _key_object = None
    _catalog_name = None

    def __init__(self, keyobject: KeyObject = None, catalog_name: str = None):
        self._key_object = keyobject
        self._catalog_name = catalog_name

    def create_request(self, header=None) -> str:
        request = 'create keyobject {} in catalog {}'.format(self._key_object.serialize(), self._catalog_name)
        logging.debug('create_keyobject_request request: ' + request)
        return request

class KeySqlCreateStoreStatement(KeySqlStatement):
    """Create Store Statement
    String query equivalent:
    'create store <store_name> for catalog <catalog_name>'
    """
    _store_name = None
    _catalog_name = None
    _owner_name = None

    def __init__(self, store_name: str=None, catalog_name: str=None, owner: str=None) -> None:
    #def __init__(self, store_name: str, catalog_name: str, schema_name: str="", owner_name: str= "", host_names: list=[], hosts = [], component_names = [], constraints = []) -> None:
        self._store_name = store_name
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        return 'create store {} for catalog {}'.format(self._store_name, self._catalog_name)


class ShowKeyObjectStatement(KeySqlStatement):
    """Show KeyObject Statement
       Connection executes showKeyObjectStatement as:
       connection.execute()
    """
    _name = None
    _catalog_name = None

    #members
    def __init__(self, name: str = None, catalog_name: str = None):
        """Create ShowKeyObjectStatement
        :param name (str): Keyobject name, non-empty
        :param catalog_name (str): Catalog name, non-empty
        """
        self._name = name
        self._catalog_name = catalog_name


    def create_request(self) -> str:
        request = 'show keyobject {} from catalog {}'.format(self._name, self._catalog_name)
        logging.debug('show_keyobject_request request: ' + request)
        return request


    @staticmethod
    def process_keyobject_definition(ko_def, dd=OrderedDict()):
        ko_name = ko_def['keyobject_name']
        ko_type = ko_def['keyobject_type']

        if ko_name in dd:
            return  dd  #  dd[ko_name]

        if ko_type == "NUMBER":
            dd[ko_name] = KeyObjectNumber(ko_name)
        elif ko_type == "INTEGER":
            dd[ko_name] = KeyObjectInteger(ko_name)
        elif ko_type == "CHAR":
            dd[ko_name] = KeyObjectChar(ko_name)
        elif ko_type == "DATE":
            dd[ko_name] = KeyObjectDate(ko_name)

        elif ko_type == 'COMPOSITION':
            if type(ko_def['component_keyobject_names']) is list:
                comp_ko_names = [rec['keyobject_name'] for rec in ko_def['component_keyobject_names']]
            else:
                comp_ko_names = [ko_def['component_keyobject_names']['keyobject_name']]
            children = [dd[nm] for nm in comp_ko_names]
            dd[ko_name] = KeyObjectComposition(ko_name, children)

        elif ko_type == 'MULTIPLE':
            comp_ko_name = ko_def['component_keyobject_names']['keyobject_name']
            child = dd[comp_ko_name]
            dd[ko_name] = KeyObjectMultiple(ko_name, child)

        elif ko_def['component_keyobjects']:
            type_comp_keyobjects = type(ko_def['component_keyobjects'])
            if type_comp_keyobjects  is list:
                for rec in ko_def['component_keyobjects']:
                    ShowKeyObjectStatement.process_keyobject_definition(rec, dd)
            elif type_comp_keyobjects is dict:
                dd = ShowKeyObjectStatement.process_keyobject_definition(ko_def['component_keyobjects'], dd)
        else:
            dd = ShowKeyObjectStatement.process_keyobject_definition(ko_def, dd)

        return dd


    def process_reply(self, reply, header=None) -> KeyObject:
        if reply == "" or reply == []:
            return []

        if header and header['format'] not in ['json', 'keysql']:
            raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

        if header and header['format'] == 'json':
            if type(reply) is str:
                all_items = json.loads(reply)
            else:
                all_items = reply
            return all_items

        assert reply[0].instance_name() == "KEYOBJECT_INFO".lower()
        inst_val = reply[0].instance_value()
        ko_name = inst_val['keyobject_definition']['keyobject_name']

        if type(inst_val['component_keyobjects']) is dict:
            dd = ShowKeyObjectStatement.process_keyobject_definition(inst_val['component_keyobjects']['keyobject_definition'], OrderedDict())
            dd = ShowKeyObjectStatement.process_keyobject_definition(inst_val['keyobject_definition'], dd)
        elif type(inst_val['component_keyobjects']) is list:
            dd = OrderedDict()
            for keyobject_def in inst_val['component_keyobjects']:
                dd = ShowKeyObjectStatement.process_keyobject_definition(keyobject_def['keyobject_definition'], dd)
            dd = ShowKeyObjectStatement.process_keyobject_definition(inst_val['keyobject_definition'], dd)
        else:
            dd = ShowKeyObjectStatement.process_keyobject_definition(inst_val['keyobject_definition'])

        return dd[ko_name]


class DescribeKeyObjectStatement(KeySqlStatement):
    """Describe KeyObject Statement
       Connection executes describeKeyObjectStatement as:
       connection.execute()
       equivalent to: show keyobject <keyobject_name> from catalog <catalog_name>
    """
    _name = None
    _catalog_name = None

    #members
    def __init__(self, name: str = None, catalog_name: str = None):
        """Create ShowKeyObjectStatement
        :param name (str): Keyobject name, non-empty
        :param catalog_name (str): Catalog name, non-empty
        """
        self._name = name
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        request = 'describe keyobject {} from catalog {}'.format(self._name, self._catalog_name)
        logging.debug('describe_keyobject_request request: ' + request)
        return request

class DescribeKeyObjectsStatement(KeySqlStatement):
    """Describe KeyObjects Statement
       Connection executes describeKeyObjectsStatement as:
       connection.execute()
       equivalent to: show keyobjects from catalog <catalog_name>
    """
    _catalog_name = None

    #members
    def __init__(self, catalog_name: str = None):
        """Create ShowKeyObjectStatement
        :param catalog_name (str): Catalog name, non-empty
        """
        self._catalog_name = catalog_name

    def create_request(self) -> str:
        request = 'describe keyobjects from catalog {}'.format(self._catalog_name)
        logging.debug('describe_keyobjects_request request: ' + request)
        return request


class RenameKeyObjectStatement(KeySqlStatement):
    _old_name = None
    _new_name = None

    #def __init__(self, old_name: str = None, new_name: str = None, catalog: str, schema=None):
    def __init__(self, old_name: str, new_name: str, catalog: str, schema=None):

        self._old_name = old_name
        self._new_name = new_name
        self._schema = schema
        self._catalog = catalog

    def create_request(self) -> str:
        if self._schema is None:
            return 'rename KEYOBJECT {} to {} '.format(self._old_name, self._new_name)
        return 'rename KEYOBJECT {} to {} IN CATALOG {}.{}'.format(self._old_name, self._schema, self._new_name, self._catalog)

    #def process_reply(self, reply: str, header=None) -> None:
    #    pass


class AlterKeyObjectStatement(KeySqlStatement):
    _old_name = None
    _keyobject = None
    _catalog_name = None

    def __init__(self, old_name: str = None, keyobject: KeyObject = None, catalog_name: str = None):

        self._old_name = old_name
        self._keyobject = keyobject
        self._catalog_name = catalog_name

    def create_request(self):
        request = 'alter keyobject {} = {} in catalog {}'.format(self._old_name, self._keyobject.serialize(), self._catalog_name)
        logging.debug('alter_keyobject_request request: ' + request)
        return request


class DropKeyObjectStatement(KeySqlStatement):
    """
    Drop KeyObject Statement
    String query equivalent:
    'drop keyobject <keyoject_name> from catalog <catalog_full_name>'
    """
    #members
    _keyobject_name = None
    _catalog_name = None

    def __init__(self, keyobject : KeyObject = None, catalog_name : str = None):
        if keyobject is not None:
            self._keyobject_name = keyobject.__name__ # .get_name()
        else:
            self._keyobject_name = None
        self._catalog_name = catalog_name

    def create_request(self):
        request = 'drop keyobject {} from catalog {}'.format(self._keyobject_name, self._catalog_name)
        logging.debug('drop_keyobject_request request: ' + request)
        return request

class DropKeyObjectIfExistsStatement(KeySqlStatement):
    """
    Drop KeyObject If Exists Statement
    String query equivalent:
    'drop keyobject if exists <keyoject_name> from catalog <catalog_full_name>'
    """
    #members
    _keyobject_name = None
    _catalog_name = None

    def __init__(self, keyobject : KeyObject = None, catalog_name : str = None):
        if keyobject is not None:
            self._keyobject_name = keyobject.__name__ # .get_name()
        else:
            self._keyobject_name = None
        self._catalog_name = catalog_name

    def create_request(self):
        request = 'drop keyobject if exists {} from catalog {}'.format(self._keyobject_name, self._catalog_name)
        logging.debug('drop_keyobject_if_exists_request request: ' + request)
        return request


class DropKeyObjectIfExistsStatement(KeySqlStatement):
    """
    Drop KeyObject Statement
    String query equivalent:
    'drop keyobject  if exists <keyoject_name>  from catalog <catalog_full_name>'
    """
    #members
    _keyobject_name = None
    _catalog_name = None

    def __init__(self, keyobject : KeyObject = None, catalog_name : str = None):
        if keyobject is not None:
            self._keyobject_name = keyobject.__name__ # .get_name()
        else:
            self._keyobject_name = None
        self._catalog_name = catalog_name

    def create_request(self):
        request = 'drop keyobject if exists {} from catalog {}'.format(self._keyobject_name, self._catalog_name)
        logging.debug('drop_keyobject_if_exists request: ' + request)
        return request
