#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

from typing import  List
from keysql_client.keysql_store import KeySqlStore
from keysql_client.keysql_catalog_statement import *
from keysql_client.key_object_statement import CreateKeyObjectStatement,DropKeyObjectStatement,AlterKeyObjectStatement
from keysql_client.key_object_statement import DropKeyObjectStatement, DropKeyObjectIfExistsStatement
from keysql_client.keysql_sequence_statement import KeySqlDropSequenceStatement, KeySqlDropSequenceIfExistsStatement


from keysql_client.key_object import *
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage
import logging

#from keysql_client.keysql_catalog_statement import KeySqlCreateCatalogStatement

class KeySQLCatalog_STATUS(Enum):
    DO_NOTHING = 0
    TO_CREATE = 1
    TO_UPDATE = 2
    TO_DELETE = 3

class KeySqlCatalog:

    # {"CATALOG_DESCRIPTION":{"SCHEMA_NAME":"TEST3","CATALOG_NAME":"CATALOG02","STORE_NAMES":null,
    # "OWNER_NAME":"test3","KEYOBJECT_DEFINITIONS":null}}

    """KeySqlCatalog
    Keeps KeySqlObjects and stores
    """
    _dirty = None    #  DO_NOTHING:0, TO_CREATE:1, TO_UPDATE:2, TO_DELETE:3
    def __init__(self,
                 catalog_name: str,
                 schema_name="",
                 owner_name="",
                 store_names=[],
                 keyobject_definitions=[],
                 sequence_definitions=[]):
        """Create KeySqlCatalog
        :param name (str): catalog name, non-empty
        :param catalog_name (str): Catalog name, non-empty
        """
        self.__name__ = catalog_name
        self._catalog_name = catalog_name
        self._owner_name = owner_name
        self._schema_name = schema_name
        self._store_names = store_names if store_names else []
        self._keyobject_definitions = keyobject_definitions if keyobject_definitions else []
        self._sequence_definitions = sequence_definitions if sequence_definitions else []

        #TODO: get rid from ... use _store_names?  _objects -> _keyobject_definitions
        #self._objects = []  # Dict[str, KeyObject] = {}
        #self._stores = [] # {}


    def get_name(self) -> str:
        """
        catalog name
        :return: str - Keyobject name
        """
        return self.__name__


    def get_owner_name(self):
        """
        owner name
        :return: str - owner name
        """
        return self._owner_name


    def get_store_names(self):
        """
        Store names
        :return: list of store names
        """
        return self._store_names


    def get_schema_name(self):
        """
        schema name
        :return: str - schema name
        """
        return self._schema_name


    def get_keyobject_definitions(self):
        """
        Keyobject definitions
        :return: list of Keyobject definitions
        """
        return self._keyobject_definitions


    def get_sequence_definitions(self):
        """
        Sequence definitions
        :return: list of Sequence definitions
        """
        return self._sequence_definitions


    def serialize(self):
        """Serialize KeySqlCatalog class
         :param : no parameters
         :return: catalog_name (str): Catalog name, non-empty
         """
        return self.__name__


    def get_keyobject_names(self):
        """Get keyobject names
        :return: a list of keyobject names (empty list if there are no keyobjects)
        """
        if not self.get_keyobject_definitions():
            return []
        return [ko.get_name() for ko in self.get_keyobject_definitions()]


    def add_object(self, key_object: KeyObject):
        """Add KeySqlObject to catalog
        :param key_object (KeyObject):
        :return: self
        Example: add two objects to catalog
        KeySqlCatalog('catalog_name').add_object(KeyObjectChar("last_name")).add_object(KeyObjectNumber("yob"))
        """
        #self._objects[key_object.keyobject_name()] = key_object
        self._objects.append(key_object)
        return self

    def add_store(self, store : KeySqlStore):
        self._stores[store.store_name()] = store
        return self

    #def show_stores(self) -> List[str]:
    #    return KeySqlShowStoresStatement(self.__name__)

    def rename(self, name : str, schema=None) -> KeySqlStatement:
        old_name = self.__name__
        self.__name__ = name
        return KeySqlRenameCatalogStatement(old_name, self.__name__, schema=schema)

    def create(self) -> KeySqlStatement:
        return KeySqlCreateCatalogStatement(self.__name__)

    def drop(self) -> KeySqlStatement:
        return KeySqlDropCatalogStatement(self.__name__)

    def drop_if_exists(self) -> KeySqlStatement:
        return KeySqlDropCatalogIfExistsStatement(self.__name__)


    def alter(self,name : str) -> KeySqlStatement:
        old_name = self.__name__
        self.__name__ = name
        return KeySqlAlterCatalogStatement(old_name, self.__name__)

    def create_object(self,key_object: KeyObject):
        #return CreateKeyObjectStatement(keyobject=key_object, catalog_name=self.__name__)
        return CreateKeyObjectStatement(keyobject=key_object, catalog_name=self.get_name())

    def alter_object(self, name: str, keyobject : KeyObject):
        return AlterKeyObjectStatement(name, keyobject, self.__name__)

    def drop_object(self,key_object: KeyObject):
        return DropKeyObjectStatement(keyobject = key_object, catalog_name = self.__name__)

    def drop_if_exists_object(self,key_object: KeyObject):
        return DropKeyObjectIfExistsStatement(keyobject = key_object, catalog_name = self.__name__)

    def drop_sequence(self, sequence_object):
        return KeySqlDropSequenceStatement(sequence_object = sequence_object, catalog_name = self.__name__)

    def drop_if_exists_sequence(self, sequence_object):
        return KeySqlDropSequenceIfExistsStatement(sequence_object = sequence_object, catalog_name = self.__name__)

    def create_store(self, store: KeySqlStore):
        return store.create(catalog_name = self.__name__)

    def build(self) -> List[KeySqlStatement]:
        statements = []
        if self._dirty == 0:
            statements.append(self.create())

        for obj in self._objects:    # self.keyobject_create_order():
            statements.append(CreateKeyObjectStatement(obj, self.__name__))
        self._objects = []
        for store in self._stores.values():
            statements.append(store.create(self.__name__))
        self._stores = {}
        return statements


    def key_object_create_order(self) -> List[KeyObject]:
        res = []
        if len(self._objects) <= 1:
            return list(self._objects.values())

        return ExtractKeyObjectsEngine().get_keyobjects_in_order(self._objects.values())

### Helper classes ###

class ExtractKeyObjectsEngine:
    dd_processed = {}
    processed_items = []
    dd = {}

    def __init__(self):
        self.dd_processed = {}
        self.processed_items = []

    def get_keyobjects_in_order(self, all_key_objects: List[KeyObject]):
        primitive_items = []
        complex_items = []

        for item in all_key_objects:
            if type(item) in [KeyObjectChar, KeyObjectNumber, KeyObjectDate]:
                primitive_items.append(item)
            else:
                complex_items.append(item)

        for k_object in primitive_items :
            #k_object = KeyObjectBuilder().create_keysql_object_from_string(item)
            self.dd_processed[k_object.__name__] = k_object
            self.processed_items.append(k_object)

        for item in complex_items:
            self.process_complex_object(item)

        return self.processed_items

    def process_complex_object(self, sss):
        if sss.__name__ in self.dd_processed_items:
            return

        if type(sss) not in [KeyObjectMultiple, KeyObjectComposition]:
            raise KeySqlException(-1, KeySqlExceptionMessage.NO_KEYOBJECT_DEFINITION)

        if type(sss) == KeyObjectMultiple:
            if sss.__name__ not in self.processed_items:
                child = sss.__ref__[0]
                if child not in self.dd_processed_items:
                    self.process_complex_object(child)
                self.processed_items.append(sss)
                self.dd_processed[sss.__name__] = sss
            return

        for child in sss.__ref__:
            if child.__name__ not in self.dd_processed:
                self.process_object(child)
        self.dd_processed[sss.__name__] = sss
        self.processed_items.append(sss)

    def process_object_from_str(self, sss):
        start = sss.find('{')
        end = sss.find('}')
        parent = sss[:start].strip()
        if parent in self.dd:
            return

        children = sss[start + 1: end].strip()
        if children.endswith('MULTIPLE'):
            child, _ = children.split(' ')
            if child in self.dd_unprocessed:
                self.process_object(self.dd_unprocessed[child])
                del self.dd_unprocessed[child]
            k_object = KeyObjectMultiple(parent, [child])
            self.dd[k_object.__name__] = k_object
            self.processed_items.append(k_object)
        else:
            children = sss[start+1:end].strip().split(',')
            for child in children:
                if child in self.dd_unprocessed:
                    self.process_object(self.dd_unprocessed[child])
                    del self.dd_unprocessed[child]
            k_object = KeyObjectComposition(parent, children)
            self.dd[k_object.__name__] = k_object
            self.processed_items.append(k_object)

    def get_keyobjects_from_str_reply(self, reply11):
        processed_items = []

        primitive_items = [item for item in reply11 if '{' not in item]
        complex_items = [item for item in reply11 if '{' in item]

        logging.info(primitive_items)
        logging.info(complex_items)

        for item in primitive_items :
            k_object = KeyObjectBuilder().create_keysql_object_from_string(item)
            self.dd[k_object.__name__] = k_object

            if '__dict__' in dir(k_object):
                logging.info(vars(k_object))
                logging.info(k_object.serialize())

            processed_items.append(k_object)

        logging.info(self.dd)
        logging.info('total primitive key_objects:', len(processed_items))
        logging.info('total complex key_objects:  ', len(complex_items))

        dd_unprocessed = {}
        for sss in complex_items:
            start = sss.find('{')
            end = sss.find('}')
            parent = sss[:start].strip()
            dd_unprocessed[parent] = sss

        for item in complex_items:
            self.process_object(item)

        return self.processed_items

if __name__ == '__main__':
    statement = KeySqlCatalog('catalog_name').add_object(KeyObjectChar("last_name")).add_object(KeyObjectNumber("yob"))
