#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

import socket
import ssl
import tempfile
import os
import logging
from tornado import concurrent
from tornado.ioloop import IOLoop
from tornado.tcpclient import TCPClient
from tornado.gen import convert_yielded
from tornado.gen import coroutine

import locale
locale.setlocale(locale.LC_ALL, 'en_US.utf-8')

import tracemalloc

# call: StatementType(4)
from keysql_client.keysql_exception import KeySqlException, KeySqlConnectException, KeySqlExceptionMessage
from keysql_client.keysql_message_header import MessageHeader


class KeySqlConnectorNonBlocking:
    pem = ["-----BEGIN CERTIFICATE-----\n",
              "MIICKzCCAZQCCQDFtaRb2MBPLTANBgkqhkiG9w0BAQsFADBaMQswCQYDVQQGEwJV\n",
              "UzELMAkGA1UECAwCQ0ExETAPBgNVBAcMCE1pbHBpdGFzMQ0wCwYDVQQKDAR0ZXN0\n",
              "MQ0wCwYDVQQLDAR0ZXN0MQ0wCwYDVQQDDAR0ZXN0MB4XDTE5MDkxMDA0NDI1MVoX\n"
              "DTI5MDkwNzA0NDI1MVowWjELMAkGA1UEBhMCVVMxCzAJBgNVBAgMAkNBMREwDwYD\n"
              "VQQHDAhNaWxwaXRhczENMAsGA1UECgwEdGVzdDENMAsGA1UECwwEdGVzdDENMAsG\n"
              "A1UEAwwEdGVzdDCBnzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEA013a3IUOihH9\n"
              "g5BvTrVtkHsq3utFgSXfQzblaN90BjLHmEisdbaldbPTT6TXvKn3v9C7J9hnRa2h\n",
              "fGCArVvyUMF6qNcCezre4ESskD0Z7Blec2Dc0uR70QCDFIoOyYjeja7p4sAumYYe\n",
              "UYt7evCNFLq2kPse/YfbQC9Jpa0oAykCAwEAATANBgkqhkiG9w0BAQsFAAOBgQA6\n",
              "FlMV6JUtGnKnVFNN+jCjILu45kSYThMMntNCqtpcJO99U/2Ng1qQdss+TL4vfx/l\n",
              "y1aNxxydCiD2XZKkjsw1RR924u8sbqzwpnlYmaWBb3JKJ5VuhagjpwgNrdpZr/Sx\n",
              "FpdNklc47mX3SpCgYcBDnssGCKjUs83GBcQYRV+oZg==\n",
              "-----END CERTIFICATE-----\n"]

    # BaseIOStream.write(data: Union[bytes, memoryview]) → Future[None]
    # BaseIOStream.read_bytes(num_bytes: int, partial: bool = False) → Awaitable[bytes]


    def __init__(self, host: str, port:int, ioLoop:IOLoop, timeout:int = None):
        try:
            self._host = socket.gethostbyname(host)
        except ValueError as ex:
            logging.exception("message")
            raise KeySqlConnectException(ex)
        self._host = host
        self._port = port
        self._ioLoop = ioLoop
        self._server_address = (self._host, self._port)


        #self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)     # blocking
        self._stream  = None                                                   # non-blocking

        self._timeout = timeout
        self._connected = False
        self._msg_id = 0
        self._stream = None
        self._reply = None
        self._reply_header = None
        self._page = None
        self._header_str = None

    def is_connected(self):
        """
        Check if there is connection with server
        :return: None
        """
        return self._connected

    async def connect_user(self, user, password):
        reply = None
        try:
            connect_msg = bytes("connect " + user + " " + password + "\n",encoding="utf-8")
            self._stream = await TCPClient().connect(self._host, self._port)
            self._fd = self._stream.fileno()
            await self._stream.write(connect_msg)
            reply = await self._stream.read_until(b"\n")
            reply = reply.decode("ascii")
        except ValueError as ex:   #   Exception as ex:
            future = concurrent.Future()
            future.set_exception(KeySqlConnectException(ex))  # .message)
            return future

        if reply != "OK\n":
            future = concurrent.Future()
            future.set_exception(KeySqlConnectException(reply))
            return future
        self._connected = True
        return True

    # chunk = self._sock.recv(8192) # blocking code
    #chunk = await self._stream.read_bytes(8192, partial=True)  # non-blocking code
    # self._sock.sendall(message.encode())         #  blocking code
    # await self._stream.write(message.encode())    # non-blocking code

    async def __API_call(self,command: str,expected: str):
        reply = None
        try:
            #self._sock.sendall(bytes(command + "\n",encoding="utf-8"))        #  blocking code
            await self._stream.write(bytes(command + "\n",encoding="utf-8"))  # non-blocking code

            #reply = self._sock.recv(128)                      #  blocking code
            reply = await self._stream.read_bytes(128, partial=True)  # non-blocking code

            pos = reply.find(b'\n')
            reply = reply[0:pos].decode("ascii")
        except ValueError as ex:   #   Exception as ex:
            raise KeySqlConnectException(ex)  # .message)

        if reply == expected:
            return True
        raise KeySqlConnectException(reply)  # .message)

    def disconnect_user(self):
        return self.__API_call("disconnect","OK")


    def abort_query(self):
        return self.__API_call("abort","OK")

    def ping(self):
        return self.__API_call("ping","PONG")


    def _connect(self, user: str = None, password: str = None):  # , nonblocking=True):
        """
        Connect with server
        :param user: user name
        :param password: user password
        :return: None
        """

        if not self._sock_connected:
            try:
                self._sock.connect(self._server_address)

                # sock.setblocking(True) is equivalent to sock.settimeout(None)
                # sock.setblocking(False) is equivalent to sock.settimeout(0.0)
                self._sock.setblocking(False)
                #if nonblocking:
                #    self._sock.setblocking(0)

                self._sock_connected = True
            except ValueError as ex:  # Exception as ex:
                raise KeySqlConnectException(ex)  # .message)

        if user and password:
            return self.connect_user(user, password)
        return True

    def connect_user_blocked(self, user, password):
        reply = None
        try:
            init_msg = "connect " + user + " " + password + "\n"
            self._sock.sendall(bytes(init_msg,encoding="utf-8"))
            reply = self._sock.recv(128)
            pos = reply.find(b'\n')
            reply = reply[0:pos].decode("ascii")
        except ValueError as ex:   #   Exception as ex:
            raise KeySqlConnectException(ex)  # .message)

        if reply != "OK":
            raise KeySqlConnectException(reply)
        self._connected = True
        return True

    def ssl_connect(self, user: str = None, password: str = None): # , nonblocking=True):
        """
        SSL connection with server
        :param user: user name
        :param password:  user password
        :return: None
        """
        if not self._sock_connected:
            try:
                context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
                fd, tmp_path = tempfile.mkstemp()
                with open(tmp_path, 'w') as f:
                    for s in self.pem:
                        f.write(s)
                os.close(fd)
                context.load_verify_locations(tmp_path)
                self._sock = context.wrap_socket(self._sock, server_hostname = self._host)
                self._sock.connect(self._server_address)
                #self._sock.setblocking(0)
                #if nonblocking:
                #    self._sock.setblocking(0)
                #os.remove(tmp_path)
                self._sock_connected = True

            except ValueError as ex:
                raise KeySqlConnectException(ex)

        if user and password:
            return self.connect_user_blocked(user,password)
        return True

    def init_reply(self):
        self._reply = ""
        self._page = b""
        self._reply_header = None
        self._header_str = ""

    def cb_factory(self, callback, on_error=None):
        """ Returns a function that handles the ioloop call back """

        def cb(fd, ev):

            def get_header(chunk):
                pos = chunk.find(b'\n')
                if pos > 0:
                    self._header_str += chunk[0:pos].decode("ascii")
                    self._reply_header = MessageHeader(self._msg_id, "keysql").unpack(self._header_str)
                    self._header_str = ""
                    return chunk[pos + 1:]

                self._header_str += chunk.decode("ascii")
                self._reply_header = None
                return b""

            try:
                while True:
                    chunk = self._sock.recv(8192)
                    if self._reply_header is None:
                        chunk = get_header(chunk)
                    self._page += chunk
                    if self._reply_header is not None and len(self._page) >= self._reply_header["message_length"]:
                        self._reply += self._page[0:self._reply_header["message_length"]].decode("utf-8")
                        if self._reply_header["page"] == 0:
                            break
                        self._page = get_header(self._page[self._reply_header["message_length"] + 1:])
                if self._reply_header["page"] == 0:
                    self.nb_cleanup()
                    if self._reply_header['exit_code'] != 0:
                        logging.error('exit_code: ' + str(self._reply_header['exit_code']) + ', reply: ' + self._reply)
                        self._reply_header["err_message"] = self._reply
                        raise KeySqlException(self._reply_header['exit_code'], self._reply)
                    callback(self._reply_header, self._reply)
            except Exception as e:
                if on_error:
                    return on_error(e)
                else:
                    raise e
            finally:
                pass

        return cb


    def send(self, msg, callback, on_error=None, format='keysql'):
        def utf8len(s):
            return len(s.encode('utf-8'))

        if format not in ['keysql', 'json']:
            raise KeySqlException(-1, KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

        self._msg_id += 1
        header = MessageHeader(self._msg_id, format)
        message = "submit " + header.out_msg(utf8len(msg))
        message += msg
        self._sock.sendall(message.encode())
        self.init_reply()
        IOLoop.instance().add_handler(self._sock.fileno(), self.cb_factory(callback, on_error), IOLoop.READ)


    async def send_resv(self, msg, on_error=None, format='keysql'):

        """
        Send a query to the server
        :param msg: str
        :return: None
        """

        def utf8len(s):
            return len(s.encode('utf-8'))

        if format not in ['keysql', 'json']:
            raise KeySqlException(-1, KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

        self._msg_id += 1
        header = MessageHeader(self._msg_id, format)
        message = "submit " + header.out_msg(utf8len(msg))
        message += msg

        #self._sock.sendall(message.encode())         #  locking code - to compare
        await self._stream.write(message.encode())    # non-blocking code

        reply = ""
        reply_header = None
        header_str = ""
        page = b""

        def get_header(chunk,header_str):
            pos = chunk.find(b'\n')
            if pos > 0:
                header_str += chunk[0:pos].decode("ascii")
                reply_header = header.unpack(header_str)
                return chunk[pos + 1:], "", reply_header

            header_str += chunk.decode("ascii")
            return b"", header_str, None

        while True:
            #chunk = self._sock.recv(8192) # blocking code
            chunk = await self._stream.read_bytes(8192, partial=True) # non-blocking code

            if reply_header is None:
                chunk, header_str, reply_header = get_header(chunk, header_str)
            page += chunk
            if len(page) >= reply_header["message_length"]:
                reply += page[0:reply_header["message_length"]].decode("utf-8")
                if reply_header["page"] == 0:
                    break
                page, header_str, reply_header = get_header(page[reply_header["message_length"] + 1:],header_str)

        if reply_header['exit_code'] != 0:
            logging.error('exit_code: ' + str(reply_header['exit_code']) + ', reply: ' + reply)
            reply_header["err_message"] = reply
            raise KeySqlException(reply_header['exit_code'], reply)

        return reply_header, reply


    def nb_cleanup(self):
        self._ioLoop.remove_handler(self._sock.fileno())

    @coroutine
    def _disconnect(self):
        """
        Disconnect with server
        :return: None
        """
        #time.sleep(0.5)     # 5 milliseconds
        #self._sock.close()   # blocking
        self._stream.close()  # non-blocking

        self._connected = False

# to check working -------------------------------------------------------------------
async def conn1_async(my_local_host, my_port, my_user, my_pass, ioLoop):
    connector = KeySqlConnectorNonBlocking(my_local_host, my_port, ioLoop)
    await connector.connect_user(my_user, my_pass)
    print('connector: ', connector.is_connected())

    for ii in range(50):
        header, reply = None, None
        header, reply = await connector.send_resv('select * from world_bank_store')   # ('show catalogs')
        if reply:
            print("conn1_async: ", ii, '-'*50)
            print(reply[:4], ' ...')

    await connector.disconnect_user()
    return "1st finished"


async def conn2_async(my_local_host, my_port, my_user, my_pass, ioLoop):
    connector2 = KeySqlConnectorNonBlocking(my_local_host, my_port, ioLoop)
    await connector2.connect_user(my_user, my_pass)
    #print('connector2:', connector2.is_connected())

    for ii2 in range(50):
        header, reply = None, None
        header, reply = await connector2.send_resv('select * from world_bank_store')  # billiard_store')
        if reply:
            print("conn2_async: ", ii2, '='*50)
            print(reply[:38], ' ...')

    await connector2.disconnect_user()
    return "2nd finished"

def callback(f):
    print(f.result())

def two_conn(my_local_host, my_port, my_user, my_pass, ioLoop):
    f1 = convert_yielded(conn1_async(my_local_host, my_port, my_user, my_pass, ioLoop))
    #ioLoop.add_future(f1, callback)
    f2 = convert_yielded(conn2_async(my_local_host, my_port, my_user, my_pass, ioLoop))
    #ioLoop.add_future(f2, callback)


async def conn1_conn2_async(my_local_host, my_port, my_user, my_pass, ioLoop):
    connector = KeySqlConnectorNonBlocking(my_local_host, my_port, ioLoop)
    await connector.connect_user(my_user, my_pass)
    #print('connector: ', connector.is_connected())

    connector2 = KeySqlConnectorNonBlocking(my_local_host, my_port, ioLoop)
    await connector2.connect_user(my_user, my_pass)
    print('connector2:', connector2.is_connected())

    for ii2 in range(5):
        #header, reply = await connector.send_resv('show stores')
        #print("conn1_async: ", ii2, '-'*50)
        #print(reply[:4], ' ...')

        header, reply = await connector2.send_resv('select * from world_bank_store') #('select * office_rental')  # from billiard_store')
        print("conn2_async: ", ii2, '='*50)
        print(reply[:38], ' ...')

    await connector.disconnect_user()
    await connector2.disconnect_user()



if __name__ == "__main__":
    pass
    #for testing
    """
    tracemalloc.start()

    my_local_host = 'my_ip_test'   #  change to real ip number 'a.b.c.d'
    #my_port = 6666   # SSL port
    my_port = 5556   # not ssl port

    my_user = "test3"
    my_pass = "test"

    print('\n both.....')

    ioLoop = IOLoop.current()
    two_conn(my_local_host, my_port, my_user, my_pass, ioLoop)
    ioLoop.start()
    """

