from typing import List
import logging
from keysql_client.key_object import KeyObject
from keysql_client.key_object import KeyObjectInteger, KeyObjectNumber, KeyObjectChar, KeyObjectDate
from keysql_client.key_object import KeyObjectMultiple, KeyObjectComposition
import re

class PackedObject(object):
    def __init__(self, name: str):
        self.name = name

    def read_instance(self, istart: int, inp_str: str):
        value, pos = self.read_value(istart, inp_str)
        return {self.name.lower(): value}, pos

    def end_value(self, istart: int, inp_str: str):
        return istart == len(inp_str) or inp_str[istart] in [',' ,'}']

    def read_value_str(self, istart, inp_str):
        pos = istart
        while pos < len(inp_str) and not self.end_value(pos, inp_str):
            pos += 1
        if pos == istart:
            return None, pos
        return inp_str[istart: pos], pos
    
    def read_char_value(self, istart, inp_str):
        pos = istart
        if self.end_value(pos,inp_str):
            return None, pos
        tmp_start_pos_read = inp_str[pos]

        try:
            assert inp_str[pos] in ["'", "\'"]
        except:
            logging.error('assert 1: pos {}, value {}'.format(pos, inp_str[pos]))

        pos += 1
        while pos < len(inp_str):
            if inp_str[pos] == "'":
                if pos < len(inp_str) and inp_str[pos + 1] == "'":
                    pos += 1
                else:
                    break
            pos += 1

        tmp_last_pos_read = inp_str[pos]

        try:
            assert inp_str[pos] in ["'", "\'"]
        except:
            logging.error('assert 2: pos {}, value {}'.format(pos, inp_str[pos]))

        pos += 1

        s = inp_str[istart : pos]
        s2 = re.sub("''", "'", s)
        return s2, pos

class PackedInteger(PackedObject):
    def __init__(self, name: str):
        super().__init__(name)

    def read_value(self, istart, inp_str):
        temp, pos = self.read_value_str(istart, inp_str)
        if temp == None:
            return None, pos
        #print(inp_str, istart, pos, inp_str, temp, '----temp')
        return int(temp), pos

class PackedNumber(PackedObject):
    def __init__(self, name: str):
        super().__init__(name)

    def read_value(self, istart, inp_str):
        temp, pos = self.read_value_str(istart, inp_str)
        if temp == None:
            return None, pos
        return float(temp), pos

class PackedChar(PackedObject):
    def __init__(self, name: str):
        super().__init__(name)

    def read_value(self, istart, inp_str):
        return self.read_char_value(istart, inp_str)

class PackedDate(PackedObject):
    def __init__(self, name: str):
        super().__init__(name)

    def read_value(self, istart, inp_str):
        return self.read_char_value(istart, inp_str)

class PackedMultiple(PackedObject):
    __ref: PackedObject

    def __init__(self, name: str, ref: KeyObject):
        #PackedObject.__init__(self, name)
        super().__init__(name)
        self.__ref = ref

    def read_value(self, istart, inp_str):
        #print('read valulue in multiple: istart:', istart, 'inp_str: len:', len(inp_str), "={}=".format(inp_str))
        if len(inp_str) == 0:
            istart += 1
            return None, istart

        if inp_str[istart] != '{':
            return None, istart
        childs = []
        pos = istart + 1
        while pos < len(inp_str):
            inst, pos = self.__ref.read_instance(pos,inp_str)
            childs.append(inst)
            pos += 1
            if inp_str[pos - 1] == '}':
                break
        return childs, pos

class PackedComposition(PackedObject):
    __children: List[PackedObject]

    def __init__(self, name: str, children : List[PackedObject]):
        #PackedObject.__init__(self, name)
        super().__init__(name)
        self.__children = children

    def read_value(self, istart, inp_str):
        if inp_str[istart] != '{':
            return None, istart

        childs = {}
        pos = istart + 1

        for obj in self.__children:
            val, pos = obj.read_value(pos, inp_str)
            childs[obj.name.lower()] = val
            pos += 1

        return childs, pos

class KeySqlParserPackedFormat(object):

    def __init__(self):
        self.object_def = []
        self.keyobject_instanses = []

    def process_keyobject_definition(self, inp_id: int, inp_str: str):
        #print('def: ', inp_id, inp_str, len(inp_str))

        assert len(self.object_def) == inp_id
        istart = 0
        obj_def, istart = self.read_object(istart,inp_str)

        self.object_def.append(obj_def)

    def read_name(self, istart:int,  inp_str: str):
        for i in range(istart, len(inp_str)):
            if inp_str[i] in ['{', ' ']:
                return inp_str[istart:i], i

        try:
            assert i == len(inp_str)
        except:
            logging.error('assert 3: i {}, len(inpstr) {}'.format(i, len(inp_str)))


    def read_simple(self, istart:int,  inp_str: str, name: str):
        if inp_str[istart] == 'I':
            return PackedInteger(name), istart + len('INTEGER')
        if inp_str[istart] == 'C':
            return PackedChar(name), istart + len('CHAR')
        if inp_str[istart] == 'N':
            return PackedNumber(name), istart + len('NUMBER')
        return PackedDate(name), istart + len('DATE')

    def read_complex(self, istart: int, inp_str: str, name: str):
        children =  [] # List[PackedObject]
        obj, istart = self.read_object(istart,inp_str)
        children.append(obj)
        #print(children)
        #print('inp_str: ', inp_str,  "istart: ", istart, "--- in read_complex")  #  inp_str[istart])
        while istart < len(inp_str) and inp_str[istart] == ',':
            istart += 1
            obj, istart = self.read_object(istart, inp_str)
            children.append(obj)

        if len(children) == 1 and inp_str[istart] == ' ':
            obj = PackedMultiple(name,children[0])
            istart += len(' MULTIPLE')
        else:
            obj = PackedComposition(name, children)

        try:
            assert inp_str[istart] == '}'
        except:
            logging.error('assert 4: istart {}, value {}, should be }'.format(istart, inp_str[istart]))

        return obj, istart + 1

    def read_object(self, istart: int, inp_str: str):
        name, istart1 = self.read_name(istart, inp_str)
        delim = inp_str[istart1]
        istart1 += 1
        if delim == ' ':
            return self.read_simple(istart1, inp_str, name)

        try:
            assert delim == '{'
        except:
            logging.error('assert 5: delim {}, should be }'.format(delim))

        return self.read_complex(istart1, inp_str, name)

    """
    d0:{PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}}
    d1:{#PERSONS{PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}} MULTIPLE}
    d2:ID INTEGER
    """

    def process_keyobject_instance(self, inp_id: int, inp_str: str):
        ko = self.object_def[inp_id]
        istart = 0
        inst, pos = ko.read_instance(istart, inp_str)
        if len(inp_str) > 0:
            try:
                assert pos == len(inp_str)
            except:
                logging.error('assert 6: pos {}, len(inp_str) {}'.format(pos, len(inp_str)))

        if 'RESULT'.lower() in inst:
            inst = inst['RESULT'.lower()]
        self.keyobject_instanses.append(inst)


    def process_line(self, input_str):
        # [KeyObject, KeyObjectChar, KeyObjectDate, KeyObjectComposition, KeyObjectMultiple]
        if len(input_str) == 0:
            return
        ind = input_str.index(':')
        id_str = input_str[:ind]
        inp_str = input_str[ind + 1:]
        if id_str[0] == 'd':
            inp_id = int(id_str[1:])
            self.process_keyobject_definition(inp_id, inp_str)
        else:
            inp_id = int(id_str)
            self.process_keyobject_instance(inp_id, inp_str)

    def process_select_reply(self, inp_line: str):
        def special_split(inp):
            res = []
            start, i, end = 0, 0, len(inp)
            no_ordinary = True
            no_double = True

            while i < end:
                if inp[i] == "'":
                    no_ordinary = not no_ordinary
                #if inp[i] == '"':
                #        no_double = not no_double
                elif inp[i] == '\n':  # "\n":
                    if no_ordinary and no_double and start < i:
                        res.append(inp[start: i])
                        start = i + 1
                i += 1
            if start < end:
                res.append(inp[start:end])

            return res

        lines = special_split(inp_line.strip())   # old code: strip().split('\n')

        i_failed = 0
        for line in lines:
            try:
                line = line.strip()
                self.process_line(line)
            except Exception as ex:
                i_failed += 1
                print('processing line exception=============================')
                print(line.strip())
                print('processed so far: ', len(self.keyobject_instanses))
                print("ex: ", type(ex), ex.args)
                print('=======================================================')

                if i_failed >= 10:
                    break
        if i_failed > 0:
            print('total failed:', i_failed, '----------------------------')
        return self.keyobject_instanses

    @staticmethod
    def process_all_args(*args):
        pass

"""
"d0:ID INTEGER\n0:0\n"
"d0:ID INTEGER\n0:0\n0:1\n"
"d0:D DATE\n0:'1970-01-01 00:00:00'\nd1:N NUMBER\n1:1.1\n"

"d0:C{ID INTEGER,X CHAR}\n0:{0,}\n",1);

"INSERT INTO %s.my_store INSTANCES {C:{ID:0,X:'a'}}",schema_name))
"SELECT C FROM %s.my_store",schema_name),"d0:C{ID INTEGER,X CHAR}\n0:{0,'a'}\n",1);

ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {C:NULL}",schema_name));
"SELECT C FROM %s.my_store",schema_name),"d0:C{ID INTEGER,X CHAR}\n0:\n",1);

TEST_F(PackedReplytTest, MultiReplyTest) {
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {mx:{x:'a'}}",schema_name));
	ExecSelect(str::format("SELECT * FROM %s.my_store",schema_name),"d0:MX{X CHAR MULTIPLE}\n0:{'a'}\n",1);
	ASSERT_TRUE(ExecKSQL("TRUNCATE STORE %s.my_store", schema_name));
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {mx:{x:'a',x:'b'}}",schema_name));
	ExecSelect(str::format("SELECT * FROM %s.my_store",schema_name),"d0:MX{X CHAR MULTIPLE}\n0:{'a','b'}\n",1);
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {mc:{c:{id:1}}}",schema_name));
	ExecSelect(str::format("SELECT mx FROM %s.my_store",schema_name),"d0:MX{X CHAR MULTIPLE}\n0:{'a','b'}\n",1);
	ExecSelect(str::format("SELECT mc FROM %s.my_store",schema_name),"d0:MC{C{ID INTEGER,X CHAR} MULTIPLE}\n0:{{1,}}\n",1);
	ExecSelect(str::format("SELECT * FROM %s.my_store",schema_name),
			"d0:MX{X CHAR MULTIPLE}\n0:{'a','b'}\nd1:MC{C{ID INTEGER,X CHAR} MULTIPLE}\n1:{{1,}}\n",2);
}

TEST_F(PackedReplytTest, FuncReplyTest) {
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {ID:0}",schema_name));
	ExecSelect(str::format("SELECT COUNT(*) FROM %s.my_store",schema_name),"d0:COUNT INTEGER\n0:1\n",1);
	ExecSelect(str::format("SELECT MAX(ID) FROM %s.my_store",schema_name),"d0:MAX_ID INTEGER\n0:0\n",1);
	ExecSelect(str::format("SELECT MIN(ID) FROM %s.my_store",schema_name),"d0:MIN_ID INTEGER\n0:0\n",1);
	ExecSelect(str::format("SELECT MIN(ID),MAX(ID) FROM %s.my_store",schema_name),"d0:RESULT{MIN_ID INTEGER,MAX_ID INTEGER}\n0:{0,0}\n",1);
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {ID:1}",schema_name));
	ExecSelect(str::format("SELECT MIN(ID),MAX(ID) FROM %s.my_store",schema_name),"d0:RESULT{MIN_ID INTEGER,MAX_ID INTEGER}\n0:{0,1}\n",1);
	ExecSelect(str::format("SELECT MIN(ID),MAX(ID),COUNT(*) FROM %s.my_store",schema_name),"d0:RESULT{MIN_ID INTEGER,MAX_ID INTEGER,COUNT INTEGER}\n0:{0,1,2}\n",1);
	// aggr ExecSelect(str::format("SELECT SUM(ID) FROM %s.my_store",schema_name),"d0:SUM_ID INTEGER\n0:0\n",1);
	//ExecSelect(str::format("SELECT COUNT(ID) FROM %s.my_store",schema_name),"d0:COUNT_ID INTEGER\n0:1\n",1);
}

TEST_F(PackedReplytTest, SingleProjectionReplyTest) {
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {ID:0}",schema_name));
	ExecSelect(str::format("SELECT ID AS ID FROM %s.my_store",schema_name),"d0:ID INTEGER\n0:0\n",1);
	ExecSelect(str::format("SELECT ID AS $A FROM %s.my_store",schema_name),"d0:A INTEGER\n0:0\n",1);
	ExecSelect(str::format("SELECT {ID, 1 AS $A} AS $B FROM %s.my_store",schema_name),"d0:B{ID INTEGER,A INTEGER}\n0:{0,1}\n",1);
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {mc:{c:{id:1},c:{x:'a'},c:{id:2,x:'b'}}}",schema_name));
	ExecSelect(str::format("SELECT C AS C FROM %s.my_store",schema_name),"d0:C{ID INTEGER,X CHAR}\n0:{1,}\n0:{,'a'}\n0:{2,'b'}\n",3);
	//ExecSelect(str::format("SELECT C AS C FROM %s.my_store order by id",schema_name),"d0:C{ID INTEGER,X CHAR}\n0:{,'a'}\n0:{1,}\n0:{2,'b'}\n",3);
}

TEST_F(PackedReplytTest, RegularReplyTest) {
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {C:{ID:0,X:'a'}}",schema_name));
	ExecSelect(str::format("SELECT ID,X FROM %s.my_store",schema_name),"d0:RESULT{ID INTEGER,X CHAR}\n0:{0,'a'}\n",1);
	ExecSelect(str::format("SELECT _IID,ID,X FROM %s.my_store",schema_name),"d0:RESULT{_IID INTEGER,ID INTEGER,X CHAR}\n0:{1,0,'a'}\n",1);
	ExecSelect(str::format("SELECT _IID,ID,X,_VERSION FROM %s.my_store",schema_name),"d0:RESULT{_IID INTEGER,ID INTEGER,X CHAR,_VERSION}\n0:{1,0,'a',0}\n",1);
	ASSERT_TRUE(ExecKSQL("TRUNCATE STORE %s.my_store", schema_name));
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {C:null}",schema_name));
	ExecSelect(str::format("SELECT ID,X FROM %s.my_store",schema_name),"d0:RESULT{ID INTEGER,X CHAR}\n0:{,}\n",1);
	ASSERT_TRUE(ExecKSQL("TRUNCATE STORE %s.my_store", schema_name));
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {mx:{x:'a',x:'b'}}",schema_name));
	ExecSelect(str::format("SELECT X FROM %s.my_store",schema_name),"d0:#X{X CHAR MULTIPLE}\n0:{'a','b'}\n",1);
	ExecSelect(str::format("SELECT X, _VERSION FROM %s.my_store",schema_name),"d0:RESULT{#X{X CHAR MULTIPLE},_VERSION INTEGER}\n0:{{'a','b'},0}\n",1);
	ASSERT_TRUE(ExecKSQL("TRUNCATE STORE %s.my_store", schema_name));
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {mc:{c:{id:1},c:{x:'a'}}}",schema_name));
	ExecSelect(str::format("SELECT ID,X FROM %s.my_store",schema_name),"d0:RESULT{#ID{ID INTEGER MULTIPLE},#X{X CHAR MULTIPLE}}\n0:{{1,},{,'a'}}\n",1);
	ASSERT_TRUE(ExecKSQL("TRUNCATE STORE %s.my_store", schema_name));
	ASSERT_TRUE(ExecKSQL("INSERT INTO %s.my_store INSTANCES {mc:{c:{id:1,x:'a'},c:{x:'b',id:2}}}",schema_name));
	ExecSelect(str::format("SELECT ID,X FROM %s.my_store",schema_name),"d0:RESULT{#ID{ID INTEGER MULTIPLE},#X{X CHAR MULTIPLE}}\n0:{{1,2},{'a','b'}}\n",1);
"""


if __name__ == '__main__':

    parser = KeySqlParserPackedFormat()

    inp_str = """
    d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
    """

    """
    0:{'John','Smith',,{{,'(123)456-7890'},{'email','jsmith@example.com'}}}
    d1:#PERSONS{PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}} MULTIPLE}
    1:
    0:{'Joanne',,2000,}
    d2:ID INTEGER
    0:{'John','Smith',,{{,'(123)456-7890'},{'email','jsmith@example.com'}}}
    2:123
    """

    inp_str = """
    d2:ID INTEGER
    2:123
    """

    inp_str = """
    d0:ID_INT INTEGER
    0:123
    d1:ID_NUMBER NUMBER
    1:456.456
    d2:ID_CHAR CHAR
    2:"Abc"
    """

    inp_str = """
    d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR}
    0:{'John','Smith'}
    """

    inp_str = """
    d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
    0:{'John','Smith',,{{,'(123)456-7890'},{'email','jsmith@example.com'}}}
    """

    inp_str = """
        d0:CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}
        0:{{,'(123)456-7890'},{'email','jsmith@example.com'}}
        """

    inp_str = """
    d0:ID_INT INTEGER
    0:123
    d1:ID_NUMBER NUMBER
    1:456.456
    d2:ID_CHAR CHAR
    2:"Abc"
    """

    # check newline - for special split
    inp_str = """
    d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
    0:{'Jo
    hn','Smith',,{{,'(123)456-7890'},{'email','jsmith@example.com'}}}
    """
    #res = parser.process_select_reply(inp_str)
    #print("---res insts --------------------------------------------------------------------- ---")
    #print(res)
    #assert len(res) == 1

    inp_str = """
    d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
    0:{'''Jo''hn''','""Smi""th""',,{{,'(123)456-7890'},{'email','jsmith@example.com'}}}
    """

    inp_str = """
    d0:X{KO2 INTEGER,KK{KO1 INTEGER,KO2 INTEGER}}
    0:{12,{2,12}}
    0:{,{1,}}
    0:{13,{,13}}
    """

    print("--- inp_str ---------------------------------------------------------------------------")
    print(len(inp_str))
    print("===", inp_str, "===")
    parser = KeySqlParserPackedFormat()
    res = parser.process_select_reply(inp_str)
    print("--- res insts --------------------------------------------------------------------- ---")
    print(res)

    # exit(0)

    inp_str = """
    d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
    0:{'John','Smith',,{{,'(123)456-7890'},{'email','jsmith@example.com'}}}
    d1:#PERSONS{PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}} MULTIPLE}
    1:
    0:{'Joanne',,2000,}
    
    d2:ID INTEGER
    0:{'John','Smith',,{{,'(123)456-7890'},{'email','jsmith@example.com'}}}
    2:123
    """

    # ok
    inp_str = """
     d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
     0:{'Jo"a"nne',,2000,}
     """
    # OK
    inp_str = """
     d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
     0:{'Jo'a'nne',,2000,}
     """

    inp_str = """
     d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
     0:{'Jo''a''nne',,2000,}
     """

    inp_str = """
     d0:PERSON{FIRST_NAME CHAR,LAST_NAME CHAR,YOB INTEGER,CONTACTS{CONTACT{TYPE CHAR,VALUE CHAR} MULTIPLE}}
     0:{'Joanne',,2000,}
     """

    print("=== inp str =====================================================")
    print(inp_str)
    print('inp str start, end: ', ord(inp_str[0]), ord(inp_str[-1]), "==={}==={}===".format(inp_str[0], inp_str[-1]))
    #print(""")

    inp_str = inp_str.strip()
    print("after strip.......")
    print('inp str start, end: ', ord(inp_str[0]), ord(inp_str[-1]), "==={}==={}===".format(inp_str[0], inp_str[-1]))
    print("========================================================")

    parser = KeySqlParserPackedFormat()
    res = parser.process_select_reply(inp_str)
    print("---keyobject_def_base-----------------------------------------------------------------")
    #for item in parser.keyobject_def_base:
    #    print(item.__dict__, item)
    print("---keyobject_def----------------------------------------------------------------------")
    #for item in parser.keyobject_def:
    #     print(item.__dict__, item)
    print("---res insts --------------------------------------------------------------------- ---")
    print(res)
    for item in res:
        print(type(item))
        if type(item) is not dict:
            print(item,  type(item), item.__dict__)
        else:
            print(item, type(item))

    print("---object_def----------------------------------------------------------------------")
    for item in parser.object_def:
        print(item.__dict__, item)