#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
import logging
from typing import List, Optional
from keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage


class KeySqlShowSchemaStatement(KeySqlStatement):
    """Show Schema Statement
    String query equivalent:
    'show schema <name>'
    """
    _schema_name = None

    def __init__(self, schema_name: str = None) -> None:
        self._schema_name = schema_name

    def create_request(self) -> str:
        return 'show schema {}'.format(self._schema_name)

class KeySqlListSchemasStatement(KeySqlStatement):
    """Lis Schema Statement
    String query equivalent:
    'list schemas'
    """
    def __init__(self) -> None:
        pass

    def create_request(self) -> str:
        return 'list schemas'


class KeySqlDescribeSchemaStatement(KeySqlStatement):
    """Describe Schema Statement
    String query equivalent:
    'describe schema <name>'
    """
    _schema_name = None

    def __init__(self, schema_name: str = None) -> None:
        self._schema_name = schema_name

    def create_request(self) -> str:
        return 'describe schema {}'.format(self._schema_name)


class KeySqlShowSchemasStatement(KeySqlStatement):
    """Show Schemas Statement
    String query equivalent:
    'show schemas'
    """
    _catalog = None

    def __init__(self) -> List[str]:
        """
        :param no
        """
        pass

    def create_request(self) -> str:
        return 'show schemas'

    def process_reply(self, reply, header=None) -> List:
        if reply == '':
            return []

        # {SCHEMA_NAME:'DUMMY_SCHEMAZ'},{SCHEMA_NAME:'DUMMY_SCHEMAZZ'}, ...
        if header and header['format'] == 'json':
            if type(reply) is list:
                return [rec["SCHEMA_NAME"] for rec in reply]

        # else: 'keysql' format
        #list of instances:   [ {'_name': 'SCHEMA_NAME', '_value': 'DUMMY_SCHEMAZ'}, ...]
        if header and header['format'] == 'keysql':
            if type(reply) is list:
                if reply[0].instance_name().upper() != 'SCHEMA_NAME':
                    raise KeySqlException(KeySqlExceptionMessage.INSTANCE_KEY_SCHEMA_NAME)
                return [rec.instance_value() for rec in reply]

        raise KeySqlException("__not imlemeted schema statement_______")

        #TODO: remove as obsolete
        if type(reply) is str:
            all_items_names = reply.strip()
            all_items_names = all_items_names.split(',\n')
            return all_items_names


class KeySqlDescribeSchemasStatement(KeySqlStatement):
    """Describe Schemas Statement
    String query equivalent:
    'describe schemas'
    """

    def __init__(self) -> List[str]:
        """
        :param no
        """
        pass

    def create_request(self) -> str:
        return 'describe schemas'


# TODO: not implemented in server code Not callable yet
class KeySqlShowUsersStatement(KeySqlStatement):
    """Show Users Statement
    String query equivalent:
    'show users'
    """

    def __init__(self) -> List[str]:
        """
        :param no
        """
        pass

    def create_request(self) -> str:
        return 'show users'

    def process_reply(self, reply: str, header=None) -> List:
        all_items_names = reply.strip()
        if all_items_names == '':
            return []

        if header and header['format'] == 'json':
            if type(reply) is list:
                return reply

        all_items_names = reply.strip()
        if all_items_names == '':
            return []
        all_items_names = all_items_names.split(',\n')
        return all_items_names


class KeySqlCreateSchemaStatement(KeySqlStatement):
    """Create Schema Statement
    String query equivalent:
    'create schema <schema_name>'
    """
    _schema_name = None

    def __init__(self, schema_name: str = None, storage_name: str = None) -> None:
        self._schema_name = schema_name
        self._storage_name = storage_name

    def create_request(self) -> str:
        if self._schema_name:
            if self._storage_name:
                return 'create schema {} with storage = {}'.format(self._schema_name, self._storage_name)
            else:
                return 'create schema {}'.format(self._schema_name)

class KeySqlDropSchemaStatement(KeySqlStatement):
    """
    Drop Schema Statement
    String query equivalent:
    'drop schema <schema_name>'
    """

    def __init__(self, name: str = None) -> None:
        """
        :param name (str): schema name to drop
        Example:
       KeySqlDropSchemaStatement('schema_name')
        """
        self._name = name

    def create_request(self) -> str:
        return 'drop schema {}'.format(self._name)


class KeySqlDropSchemaIfExistsStatement(KeySqlStatement):
    """
    Drop Schema Statement
    String query equivalent:
    'drop schema if exists <schema_name> '
    """

    def __init__(self, name: str = None) -> None:
        """
        :param name (str): schema name to drop
        Example:
       KeySqlDropSchemaIfExistsStatement('schema_name')
        """
        self._name = name

    def create_request(self) -> str:
        return 'drop schema if exists {}'.format(self._name)


class KeySqlAlterSchemaStatement(KeySqlStatement):
    """
    String query equivalent:
    'alter schema <schema_name> = <new_name>'
    """
    def __init__(self, schema_name: str=None, new_schema_name: str=None, new_owner: Optional[str]=None, new_storage_name: Optional[str] = None) -> None:
        """
        Alter Schema Statement
        :param schema_name (str): old schema name
        :param new_schema_name (str): new schema name, non-empty, unique
        :param new_owner (str): new owner name, if empty: the same owner, if not empty: a new existing owner
        """
        self._schema_name = schema_name
        self._new_schema_name = new_schema_name
        self._new_owner_name = new_owner
        self._new_storage_name = new_storage_name

    def create_request(self) -> str:
        """
        if not self._new_owner_name:
            if not self._new_storage_name:
                return 'alter schema {} with name = {}'.format(self._schema_name, self._new_schema_name)
            else:
                return 'alter schema {} with name = {}, storage = {}'.format(self._schema_name, self._new_schema_name, self._new_storage_name)
        else:
            if not self._new_storage_name:
                return 'alter schema {} with name = {}, owner = {}'.format(self._schema_name, self._new_schema_name, self._new_owner_name)
            else:
                return 'alter schema {} with name = {}, owner = {}, storage = {}'.format(self._schema_name, self._new_schema_name,
                                                                           self._new_owner_name, self._storage_name)
        """

        query = "alter schema {} with".format(self._schema_name)
        query_change_values = []
        if self._new_schema_name:
            query_change_values.append(" name = {}".format(self._new_schema_name))
        if self._new_owner_name:
            query_change_values.append(" owner = {}".format(self._new_owner_name))
        if self._new_storage_name:
            query_change_values.append(" storage = {}".format(self._new_storage_name))

        if len(query_change_values) == 0:
            raise KeySqlException(KeySqlExceptionMessage.ALTER_SCHEMA_HAS_NO_PATAMETERS)
        else:
            return query + ",".join(query_change_values)


class KeySqlRenameSchemaStatement(KeySqlStatement):
    """
    String query equivalent:
    'alter schema <schema_name> = <new_name>'
    """
    def __init__(self, schema_name: str=None, new_schema_name: str=None) -> None:
        """
        Alter Schema Statement
        :param schema_name (str): old schema name
        :param new_schema_name (str): new schema name, non-empty, unique
        :param new_owner (str): new owner name, if empty: the same owner, if not empty: a new existing owner
        """
        self._schema_name = schema_name
        self._new_schema_name = new_schema_name

    def create_request(self) -> str:
        return 'rename schema {} to {}'.format(self._schema_name, self._new_schema_name)


if __name__ == "__main__":

    show_schemas = KeySqlShowSchemasStatement()
    create_schema = KeySqlCreateSchemaStatement('schema_name')
    drop_schema = KeySqlDropSchemaStatement('schema_name')
    alter_schema = KeySqlAlterSchemaStatement('schema1', 'schema2')
